# This file is part of the digitalSTROM dsm-api library.
#
# Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland
#
# digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# digitalSTROM dsm-api library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 
from string import Template

class Parameter:
    @classmethod
    def getDoc(cls, parameter, isReturn=False):
        out = "@param ["
        if (isReturn):
            out += "out"
        else:
            out += "in"
        out += "] " + parameter.name + " " + parameter.description
        if (isReturn and parameter.type_count > 1):
            out += " (pass a reference to a preallocated buffer of at least " + str(parameter.type_count) + " bytes.)"
        return out
        
    @classmethod
    def signature(cls, parameter):
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "dsid_t DSID"
        else:
            out = "uint" + str(parameter.type_size) + "_t "
            if (parameter.type_count > 1):
                out += "*"
            out += parameter.name
            return out
        
    @classmethod
    def declare_by_reference(cls, parameter):
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "dsid_t *DSID"
        else:
            return "uint" + str(parameter.type_size) + "_t *" + parameter.name
        
    @classmethod
    def getFromResponse(cls, parameter):
        out = "    if (" + parameter.name + " != NULL)\n"
        out += "    {\n"
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            out += "        ResponseGetArray" + str(parameter.type_size) + "(*pFrameAnswer, " + str(parameter.offset) + ", DSID->id, " + str(parameter.type_count) + ");\n"
        elif (parameter.type_count == 1):
            out += "        *" + parameter.name + " = ResponseGetData" + str(parameter.type_size) + "(*pFrameAnswer, " + str(parameter.offset) + ");\n"
        else:
            out += "        ResponseGetArray" + str(parameter.type_size) + "(*pFrameAnswer, " + str(parameter.offset) + ", " + parameter.name + ", " + str(parameter.type_count) + ");\n"
        out += "    }\n"
        return out
        
    @classmethod
    def getFromRequest(cls, parameter):
        out = "    if (" + parameter.name + " != NULL)\n"
        out += "    {\n"
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            out += "        RequestGetArray" + str(parameter.type_size) + "(*pFrameAnswer, " + str(parameter.offset) + ", DSID->id, " + str(parameter.type_count) + ");\n"
        elif (parameter.type_count == 1):
            out += "        *" + parameter.name + " = RequestGetData" + str(parameter.type_size) + "(*pFrameAnswer, " + str(parameter.offset) + ");\n"
        else:
            out += "        RequestGetArray" + str(parameter.type_size) + "(*pFrameAnswer, " + str(parameter.offset) + ", " + parameter.name + ", " + str(parameter.type_count) + ");\n"
        out += "    }\n"
        return out
    
    @classmethod
    def setRequest(cls, parameter):
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "RequestSetArray" + str(parameter.type_size) + "(pFrame, " + str(parameter.offset) + ", DSID.id, " + str(parameter.type_count) + ");"
        elif (parameter.type_count == 1):
            return "RequestSetData" + str(parameter.type_size) + "(pFrame, " + str(parameter.offset) + ", " + parameter.name + ");"
        else:
            return "RequestSetArray" + str(parameter.type_size) + "(pFrame, " + str(parameter.offset) + ", " + parameter.name + ", " + str(parameter.type_count) + ");"
    
    @classmethod
    def setResponse(cls, parameter):
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "ResponseSetData" + str(parameter.type_size) + "(pFrame, " + str(parameter.offset) + ", DSID.id, " + str(parameter.type_count) + ");"
        elif (parameter.type_count == 1):
            return "ResponseSetData" + str(parameter.type_size) + "(pFrame, " + str(parameter.offset) + ", " + parameter.name + ");"
        else:
            return "ResponseSetArray" + str(parameter.type_size) + "(pFrame, " + str(parameter.offset) + ", " + parameter.name + ", " + str(parameter.type_count) + ");"
        
    @classmethod
    def call_as_object_by_reference(cls, parameter):
        out = ""
        if (parameter.type_count == 1 or (parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12)):
            out += "&"
        out += parameter.name
        return out
        
    @classmethod
    def call(cls, parameter):
        return parameter.name
        
    @classmethod
    def declaration(cls, parameter):
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "dsid_t DSID"
        else:
            out = "uint" + str(parameter.type_size) + "_t " + parameter.name
            if (parameter.type_count > 1):
                out += "[" + str(parameter.type_count) + "]"
            return out
    
class Call:
    @classmethod
    def documentation_call(cls, call):
        s = Template("""/**
 * $description
 *
 * This call has a timeout of $timeout seconds.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSID Destination of the call
 * $request_response
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */""")
        return s.substitute(description = call.description,
                            timeout = call.timeout,
                            request_response = "\n * ".join(map(lambda x: Parameter.getDoc(x), call.request_params) + 
                                                            map(lambda x: Parameter.getDoc(x, True), call.response_params)))

    @classmethod
    def call_signature(cls, call):
        out = "int " + call.name + "(DsmApiHandle_t handle, dsid_t dSMdSID"
        if call.request_params:
            out += ", " + ", ".join(map(lambda x: Parameter.signature(x), call.request_params))
        if call.response_params:
            out += ", " + ", ".join(map(lambda x: Parameter.declare_by_reference(x), call.response_params))
        out += ")"
        return out
    
    @classmethod
    def callback_documentation(cls, call, call_type):
        s = Template("""/**
 * $description
 *
 * Callback function prototype.
 * 
 * To register this callback function, call the #DsmApiRegisterCallback function
 * with MessageType = #${message_type}, MessageId = #$messageId and ModifierId = #$modifierId
 *
 * @param [in] error Error result of the (potential) previous call.
 * @param [in] arg User data
 * @param [in] sourceId Source address of the call
 * @param [in] destinationId Destination address of the call
 * $parameters
 */""")
        if call_type == "response":
            type_const = "DS485_CONTAINER_RESPONSE"
            parameters = call.response_params
        elif call_type == "request":
            type_const = "DS485_CONTAINER_REQUEST"
            parameters = call.request_params
        else:
            type_const = "DS485_CONTAINER_EVENT"
            parameters = call.response_params
            
        return s.substitute(description = call.description,
                            message_type = type_const,
                            parameters = "\n * ".join(map(lambda x: Parameter.getDoc(x), parameters)), 
                            messageId = call.message_id,
                            modifierId = call.modifier_id)
        
    @classmethod
    def callback_declaration(cls, call, call_type):
        if call_type == "request":
            parameters = call.request_params
        else:
            parameters = call.response_params
            
        out = "typedef void (*" + call.name + "_" + call_type + "_callback_t)(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId"
        if parameters:
            out += ", " + ", ".join(map(lambda x: Parameter.signature(x), parameters))
        out += ");"
        return out
        
    @classmethod
    def parse_method_signature(cls, call, call_type):
        if call_type == "request":
            parameters = call.request_params
        else:
            parameters = call.response_params
        
        out = "static int " + call.name + "_parse_" + call_type + "(const ds485_container_t *pFrameAnswer"
        if parameters:
            out += ", " + ", ".join(map(lambda x: Parameter.declare_by_reference(x), parameters))
        out += ")"
        return out
    
    @classmethod
    def parse_method(cls, call, call_type):
        out = cls.parse_method_signature(call, call_type) + "\n"
        out += "{\n"
        if call_type == "request":
            out += "".join(map(lambda x: Parameter.getFromRequest(x), call.request_params))
            out += "    return 0;\n"
        else:
            out += "".join(map(lambda x: Parameter.getFromResponse(x), call.response_params))
            out += "    return GetErrorCode(*pFrameAnswer);\n"
        out += "}"
        return out
        
    @classmethod
    def parse_method_call(cls, call, call_type):
        if call_type == "request":
            parameters = call.request_params
        else:
            parameters = call.response_params
        
        out = call.name + "_parse_" + call_type + "(&pFrameAnswer"
        if parameters:
            out += ", " + ", ".join(map(lambda x: Parameter.call(x), parameters))
        out += ")"
        return out
        
    @classmethod
    def sync_method(cls, call):
        s = Template("""$signature
{
    ds485_container_t pFrame;
    ds485_container_t pFrameAnswer;
    dsm_api_data_t *data = (dsm_api_data_t*)handle;
    int result;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    pFrame.containerType = DS485_CONTAINER_REQUEST;
    SetDSID(pFrame.destinationId, dSMdSID);
    SetCommand(pFrame, $message_id);
    SetModifier(pFrame, $modifier_id);
    $parameter
    SetParameterLength(pFrame, $parameter_length);

    result = ds485_client_send_sync_command(data->client, &pFrame, &pFrameAnswer, $timeout);
    if (result == 0)
    {
        return $parse_response;
    }
    else
    {
        return result;
    }
}""")
            
        return s.substitute(signature = Call.call_signature(call),
                            message_id = call.message_id,
                            modifier_id = str(call.modifier_id),
                            parameter = "\n    ".join(map(lambda x: Parameter.setRequest(x), call.request_params)),
                            parameter_length = sum(map(lambda x: x.getLength(), call.request_params)),
                            timeout = call.timeout,
                            parse_response = Call.parse_method_call(call, "response"))
        
    @classmethod
    def callback_selector(cls, call, call_type):
        if call_type == "response":
            type_const = "DS485_CONTAINER_RESPONSE"
            parameters = call.response_params
        elif call_type == "request":
            type_const = "DS485_CONTAINER_REQUEST"
            parameters = call.request_params
        else:
            type_const = "DS485_CONTAINER_EVENT"
            parameters = call.response_params
        
        s = Template("""    if (MessageType == $message_type && MessageId == ${message_id} && ModifierId == ${modifier_id})
    {
        ${param_declarations}
        error = ${call_parse_method};
        ((${name}_${call_type}_callback_t)callback_entry->callback)(error, callback_entry->arg, container->sourceId, container->destinationId${parameter_call});

        return 0;
    }""")
        return s.substitute(message_type = type_const,
                            message_id = str(call.message_id),
                            modifier_id = str(call.modifier_id),
                            name = call.name,
                            call_type = call_type,
                            param_declarations = "\n        ".join(["uint8_t error;"] + map(lambda x: Parameter.declaration(x) + ";", parameters)),
                            call_parse_method = cls.parse_method_call_by_reference(call, call_type, "container"),
                            parameter_call = "".join(map(lambda x: ", " + Parameter.call(x), parameters)))
    
    @classmethod
    def parse_method_call_by_reference(cls, call, call_type, container):
        if call_type == "request":
            parameters = call.request_params
        else:
            parameters = call.response_params
        
        out = call.name + "_parse_" + call_type + "(" + container
        if parameters:
            out += ", " + ", ".join(map(lambda x: Parameter.call_as_object_by_reference(x), parameters))
        out += ")"
        return out
        
class CGenerator:
    def __init__(self, license, calls, events):
        self.calls = calls
        self.events = events
        self.license = license
        
    def header(self, header_file):
        header = open(header_file, 'w')
        print >> header, self.license
        print >> header, """/**
 * @file dsm-api.h
 *
 * This is the dSM-API declaration.
 *
 * General remark: If any return parameters are of no interest to the caller
 *                 NULL can be passed instead of a variable reference.
 */

#ifndef DSM_API_H
#define DSM_API_H
#include <digitalSTROM/ds.h>
#include <digitalSTROM/ds485.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Type of the library handle.
 */
typedef void *DsmApiHandle_t;

/**
 * Callback function signature for the BusChange callback.
 */
typedef void (*BusChange_callback_t)(void *arg, dsid_t *id, int flag);

/**
 * Callback function signature for the BusState callback.
 */
typedef void (*BusState_callback_t)(void *arg, bus_state_t state);

/**
 * Initialize the internal memory structures of the library.
 *
 * @return a handle to the library instance
 */
DsmApiHandle_t DsmApiInitialize(void);

/**
 * Open a new connection to the dS485 bus.
 * 
 * @param [in] handle Handle to the library instance
 * @param [in] connection string containing the connection method: 
 *                        - serial:///dev/ttyUSB0 (only when linked to the GPL version of the ds485-client library)
 *                        - tcp://localhost:8442
 * @param [in] flags
 * @return zero on success, negative on failure
 */
int DsmApiOpen(DsmApiHandle_t handle, const char *connection, int flags);

/**
 * Get the dSID of the daeomon/socket interface.
 *
 * @param [in] handle Handle to the library instance
 * @param [out] dsid own dSID
 * @return 0 on success; negative on failure
 */
int DsmApiGetOwnDSID(DsmApiHandle_t handle, dsid_t *dsid);

/**
 * Request a list of the current bus members.
 * 
 * @param [in] handle Handle to the library instance
 * @param [out] device_list List of devices. Preallocate a buffer of size device_list_size * sizeof(dsid_t)
 * @param [in] device_list_size device_list has space for this many devices.
 * @return the number of devices in device_list; negative if an error occured
 */
int DsmApiGetBusMembers(DsmApiHandle_t handle, dsid_t *device_list, int device_list_size);

/**
 * Get the current state of the dS485 bus.
 *
 * @param [in] handle Handle to the library instance
 * @param [out] state returns the bus state
 * @return 0 on success; negative on failure
 */
int DsmApiGetBusState(DsmApiHandle_t handle, bus_state_t *state);

/**
 * Close the connection to the dS485 bus.
 *
 * @param [in] handle Handle to the library instance
 * @return 0 on success; negative on failure
 */
int DsmApiClose(DsmApiHandle_t handle);

/**
 * Free instance memory.
 *
 * @param [in] handle Handle to the library instance
 * @return 0 on success; negative on failure
 */
int DsmApiCleanup(DsmApiHandle_t handle);

/**
 * Register callback functions for dsm-api asynchronous responses, requests and events.
 *
 * Register NULL to delete a callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] MessageType Register callback function for this MessageType
 * @param [in] MessageId Register callback function for this MessageId
 * @param [in] ModifierId Register the callback function for thei ModifierId
 * @param [in] callback_function the callback function
 * @param [in] arg user data passed back to the called function
 * @return the previously registered callback function for this MessageId/ModifierId combination
 */
void *DsmApiRegisterCallback(DsmApiHandle_t handle, container_type_t MessageType, uint8_t MessageId, uint8_t ModifierId, void *callback_function, void *arg);

/**
 * Register a callback function for BusState events.
 *
 * Register NULL to delete the callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] callback_function the callback function
 * @param [in] arg user data passed back to the called function
 * @return the previously registered callback function
 */
void *DsmApiRegisterBusStateCallback(DsmApiHandle_t handle, BusState_callback_t callback_function, void *arg);

/**
 * Register a callback function for BusChange events.
 *
 * Register NULL to delete the callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] callback_function the callback function
 * @param [in] arg user data passed back to the called function
 * @return the previously registered callback function
 */
void *DsmApiRegisterBusChangeCallback(DsmApiHandle_t handle, BusChange_callback_t callback_function, void *arg);

/**
 * Requests statistics data of a device. Synchronous implementation of #DeviceStatisticsRequest_get.
 *
 * ParamOffset 0xFE (Reset) and 0xFF (All values) are not allowed, as they would result in no or multiple response events.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ParamOffset 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Statistics value (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceStatisticsRequest_get_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t ParamOffset, int timeout, uint16_t *Value);

/**
 * Read a configuration parameter of a specific digitalSTROM device. Synchronous implementation of #DeviceConfig_get.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ConfigClass 
 * @param [in] ConfigIndex 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter value (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceConfig_get_sync_8(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint8_t *Value);

/**
 * Read two configuration parameters of a specific digitalSTROM device simultaneously. Synchronous implementation of #DeviceConfig_get.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ConfigClass 
 * @param [in] ConfigIndex 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter values low byte: ConfigIndex & 0xFE; high byte: ConfigIndex & 0xFE + 1 (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceConfig_get_sync_16(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint16_t *Value);

/**
 * Query transmission quality. Synchronous implementation of #TestTransmissionQuality_get.
 *
 * This call has a timeout of 5 seconds.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSID Destination of the call
 * @param [in] DeviceId 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] DownstreamQuality Downstream quality (only valid if return == 0)
 * @param [out] UpstreamQuality Upstream quality (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int TestTransmissionQuality_get_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, int timeout, uint8_t *DownstreamQuality, uint16_t *UpstreamQuality);
"""
        for call in self.calls:
            print >> header, Call.documentation_call(call)
            print >> header, Call.call_signature(call) + ";\n"
            print >> header, Call.callback_documentation(call, "response")
            print >> header, Call.callback_declaration(call, "response") + "\n"
            print >> header, Call.callback_documentation(call, "request")
            print >> header, Call.callback_declaration(call, "request") + "\n"
        
        print >> header, "\n"
        for event in self.events:
            print >> header, Call.callback_documentation(event, "event")
            print >> header, Call.callback_declaration(event, "event") + "\n"
        print >> header, """#ifdef __cplusplus
}
#endif

#endif
"""
        
        header.close()
        
    def implementation(self, impl_file):
        source = open(impl_file, 'w')
        print >> source, self.license
        print >> source, """#include "dsm-api.h"
#include "dsm-api-const.h"
#include <digitalSTROM/ds485-client-interface.h>
#include "dsm-macros.h"

#include "dsm-api-static.c"
"""
        for call in self.calls:
            print >> source, Call.parse_method_signature(call, "response") + ";"
            print >> source, Call.parse_method_signature(call, "request") + ";"
        for event in self.events:
            print >> source, Call.parse_method_signature(event, "event") + ";"
        
        print >> source, "\n"
        for call in self.calls:
            print >> source, Call.parse_method(call, "response") + "\n"
            print >> source, Call.parse_method(call, "request") + "\n"
            print >> source, Call.sync_method(call) + "\n"
            
        for event in self.events:
            print >> source, Call.parse_method(event, "event") + "\n"
    
        print >> source, """static int dsm_api_ds485_message_callback(void *handle, const ds485_container_t *container)
{
    uint8_t MessageId = GetCommand(*container);
    uint8_t ModifierId = GetModifier(*container);
    container_type_t MessageType = container->containerType;
    uint32_t hash_key = (MessageType << 16) | (MessageId << 8) | (ModifierId);
    struct callback_item *callback_entry = NULL;
    
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }
    
    HASH_FIND_INT(data->callback_list, &hash_key, callback_entry);
    if (callback_entry == NULL)
    {
        return -1;
    }
    
    if (callback_entry->callback == NULL)
    {
        return -1;
    }
"""
        for call in self.calls:
            print >> source, Call.callback_selector(call, "response") + "\n"
            print >> source, Call.callback_selector(call, "request") + "\n"
        for event in self.events:
            print >> source, Call.callback_selector(event, "event") + "\n"
        print >> source, """    return -1;
}
"""
        source.close()
    
    def const_header(self, const_header_file, message_ids, error_list, length_enum):
        const = open(const_header_file, 'w')
        print >> const, self.license
        
        print >> const, "enum DsmApiMsgId {"
        for i in message_ids:
            print >> const, "    " + i.get_enum_decl() + ","
        print >> const, "};"
        
        for i in message_ids:
            print >> const, i.get_modifier_enum() + "\n"
        
        if (len(error_list) > 0):
            print >> const, "enum DsmApiErrorCodes {"
            for key, value in error_list.iteritems():
                print >> const, "    " + value.getEnumEntry()
            print >> const, "};\n"
            
        if (len(length_enum) > 0):
            print >> const, "enum DsmApiParameterLength {"
            for key, value in length_enum.iteritems():
                print >> const, "    " + key + " = " + str(value) + ","
            print >> const, "};"
        const.close()
        