# This file is part of the digitalSTROM dsm-api library.
#
# Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland
#
# digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# digitalSTROM dsm-api library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.

from PyGenerator import *

class LuaParameter(PyParameter):
    @classmethod
    def getLuaPush(cls, parameter):
        if (parameter.type_count == 1):
            return "lua_pushnumber(L, " + parameter.name + ");"
        elif parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "{ char ids[25]; convert_uint8t_2_string(" + parameter.name + ".id, ids, 24);  lua_pushlstring(L, ids, 24); }"
        else:
            return "lua_pushlstring(L, " + parameter.name + ", " + str(parameter.getLength()) + ");"
    
class LuaCall(PyCall):
    @classmethod
    def lua_callback_implementation(cls, call, type):
        if (type == "request"):
            parameters = call.request_params
        else:
            parameters = call.response_params

        ltext = """void ${name}_${type}_lua_callback(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId${parameter_decl})
{
    lua_State* L = (lua_State *) arg;
    const char* callback;
    char sourceIdS[25], destinationIdS[25];
    int numargs = 3 + ${numargs};

    convert_uint8t_2_string(sourceId.id, sourceIdS, 24);
    convert_uint8t_2_string(destinationId.id, destinationIdS, 24);

    lua_pushlightuserdata(L, (void *)&${name}_${type}_lua_callback);
    lua_gettable(L, LUA_REGISTRYINDEX);
    if (!lua_isstring(L, -1)) {
       printf("Callback for ${name}_${type}_lua_callback is unknown");
       return;
    }
    callback = lua_tostring(L, -1);
    lua_getglobal(L, callback);
    if (!lua_isfunction(L, -1)) {
        printf("Callback %s is not a function\\n", callback);
        return;
    }
    lua_pushnumber(L, error_code);
    lua_pushlstring(L, sourceIdS, 24);
    lua_pushlstring(L, destinationIdS, 24);

    ${param_push}
    
    if (lua_pcall(L, numargs, 0, 0))
         printf("Callback %s failed: %s", callback, lua_tostring(L, -1));
}
"""

        numargs = 0
        parameter_decl = ""
        param_call = ""
        param_push = ""

        if parameters:
            numargs = len(parameters)
            parameter_decl = ", " + ", ".join(map(lambda x: LuaParameter.signature(x), parameters))
            param_call = ", " + ", ".join(map(lambda x: LuaParameter.getPyCall(x), parameters))
            param_push = "\n".join(map(lambda x: LuaParameter.getLuaPush(x), parameters))

        s = Template(ltext)            
        return s.substitute(name = call.name,
                            type = type,
                            numargs = numargs,
                            parameter_decl = parameter_decl,
                            param_push = param_push,
                            param_call = param_call)

    @classmethod
    def luaCallbackRegisterSelector(cls, call, call_type):
        s = Template("""    if (MessageType == ${type_const} && MessageId == ${message_id} && ModifierId == ${modifier_id})
    {
        lua_pushlightuserdata(L, (void *)&${name}_${type}_lua_callback);
        lua_pushstring(L, func);
        lua_settable(L, LUA_REGISTRYINDEX);
        DsmApiRegisterCallback(handle, MessageType, MessageId, ModifierId, &${name}_${type}_lua_callback, (void *) L);
        return;
    }""")
        
        if call_type == "response":
            type_const = "DS485_CONTAINER_RESPONSE"
        elif call_type == "request":
            type_const = "DS485_CONTAINER_REQUEST"
        else:
            type_const = "DS485_CONTAINER_EVENT"
        out = s.substitute(message_id = str(call.message_id),
                           modifier_id = str(call.modifier_id),
                           name = call.name,
                           type = call_type,
                           type_const = type_const)
        return out

class LuaGenerator:
    def __init__(self, license, calls, events):
        self.calls = calls
        self.events = events
        self.license = license
        
    def swig(self, swig_decl_file, length_enum):
        swig = open(swig_decl_file, 'w')
        print >> swig, self.license
        print >> swig, """%module dsmapi2
%{
#include "dsm-api-const.h"
#include <digitalSTROM/ds.h>
#include "stdint.h"

static void printhex(unsigned char *p, int len)
{
  int n;
  for (n = 0; n < len;n++)
    printf("%02X", p[n]);
}
static unsigned char char2hex(unsigned char c)
{
  if (c >= '0' && c <= '9')
    return c - '0';
  if (c >= 'a' && c <= 'f')
    return c - 'a' + 10;
  if (c >= 'A' && c <= 'F')
    return c - 'A' + 10;
  return 0;
}
static void convert_string_2_uint8t(const char* in, uint8_t* out, int len)
{
  int n;
  memset(out, 0, len);
  for (n = 0; n < len; n++, in+=2) {
    out[n] = char2hex(in[1]) | char2hex(in[0]) << 4;
  }
}
static void convert_uint8t_2_string(const uint8_t* in, char *out, int len)
{
  int n;
  memset(out, 0, len);
  for (n = 0; n < len; n += 2, in++) {
    out += sprintf(out, "%02X", *in);
  }
}

/* Includes the header in the wrapper code */
#include "dsm-api.h"
"""
        for call in self.calls:
            print >> swig, LuaCall.lua_callback_implementation(call, "request")
            print >> swig, LuaCall.lua_callback_implementation(call, "response")
        for event in self.events:
            print >> swig, LuaCall.lua_callback_implementation(event, "event")
            
        print >> swig, """
static void BusChange_callback(void *arg, dsid_t *id, int flag)
{
    lua_State* L = (lua_State *) arg;
    const char* callback;
    char dsids[25];

    convert_uint8t_2_string(id->id, dsids, 24);

    lua_pushlightuserdata(L, (void *)&BusChange_callback);
    lua_gettable(L, LUA_REGISTRYINDEX);
    if (!lua_isstring(L, -1)) {
       printf("Callback for BusChange_callback is registered but unknown");
       return;
    }
    callback = lua_tostring(L, -1);
    lua_getglobal(L, callback);
    if (!lua_isfunction(L, -1))
        printf("WARNING: Callback %s is not a function\\n", callback);
    lua_pushlstring(L, dsids, 24);
    lua_pushnumber(L, flag);

    if (lua_pcall(L, 2, 0, 0) != 0)
         printf("Callback %s failed: %s", callback, lua_tostring(L, -1));
}

void DsmApiRegisterLuaBusChangeCallback(DsmApiHandle_t handle, const char *func, lua_State* Lmain)
{
    static lua_State* L = NULL;
    if (L == NULL) {
        L = lua_newthread(Lmain);
        /* create a hard reference to the thread object into the registry */
        lua_pushlightuserdata(Lmain, L);
        lua_insert(Lmain, -2);
        lua_settable(Lmain, LUA_REGISTRYINDEX);
        /* move args table and function to child's stack */
        lua_xmove(Lmain, L, 1);
        lua_xmove(Lmain, L, 1);
    }

    if (func == NULL || *func == 0)
    {
        DsmApiRegisterBusChangeCallback(handle, 0, 0);
        return;
    }

    lua_getglobal(L, func);
    if (!lua_isfunction(L, -1))
        printf("WARNING: Callback %s is not a function\\n", func);
    lua_pop(L, 1);
    lua_pushlightuserdata(L, (void *)&BusChange_callback);
    lua_pushstring(L, func);
    lua_settable(L, LUA_REGISTRYINDEX);
    DsmApiRegisterBusChangeCallback(handle, &BusChange_callback, (void *) L);
}

static void BusState_callback(void *arg, bus_state_t state)
{
    lua_State* L = (lua_State *) arg;
    const char* callback;

    lua_pushlightuserdata(L, (void *)&BusState_callback);
    lua_gettable(L, LUA_REGISTRYINDEX);
    if (!lua_isstring(L, -1)) {
       printf("Callback for BusState_callback is registered but unknown");
       return;
    }
    callback = lua_tostring(L, -1);
    lua_getglobal(L, callback);
    if (!lua_isfunction(L, -1))
        printf("Callback %s is not a function\\n", callback);
    lua_pushnumber(L, (int) state);
    if (lua_pcall(L, 1, 0, 0) != 0)
         printf("Callback %s failed: %s", callback, lua_tostring(L, -1));
}

void DsmApiRegisterLuaBusStateCallback(DsmApiHandle_t handle, const char *func, lua_State* Lmain)
{
    static lua_State* L = NULL;
    if (L == NULL) {
        L = lua_newthread(Lmain);
        /* create a hard reference to the thread object into the registry */
        lua_pushlightuserdata(Lmain, L);
        lua_insert(Lmain, -2);
        lua_settable(Lmain, LUA_REGISTRYINDEX);
        /* move args table and function to child's stack */
        lua_xmove(Lmain, L, 1);
        lua_xmove(Lmain, L, 1);
    }

    if (func == NULL || *func == 0)
    {
        DsmApiRegisterBusStateCallback(handle, 0, 0);
        return;
    }

    lua_getglobal(L, func);
    if (!lua_isfunction(L, -1))
        printf("Callback %s is not a function\\n", func);
    lua_pop(L, 1);
    lua_pushlightuserdata(L, (void *)&BusState_callback);
    lua_pushstring(L, func);
    lua_settable(L, LUA_REGISTRYINDEX);
    DsmApiRegisterBusStateCallback(handle, &BusState_callback, (void *) L);
}

void DsmApiRegisterLuaCallback(DsmApiHandle_t handle, container_type_t MessageType, int MessageId, int ModifierId, const char *func, lua_State* Lmain)
{
    static lua_State* L = NULL;
    if (L == NULL) {
        L = lua_newthread(Lmain);
        /* create a hard reference to the thread object into the registry */
        lua_pushlightuserdata(Lmain, L);
        lua_insert(Lmain, -2);
        lua_settable(Lmain, LUA_REGISTRYINDEX);
        /* move args table and function to child's stack */
        lua_xmove(Lmain, L, 1);
        lua_xmove(Lmain, L, 1);
    }
    if (func == NULL || *func == 0)
    {
        DsmApiRegisterCallback(handle, MessageType, MessageId, ModifierId, 0, 0);
        return;
    }
    lua_getglobal(L, func);
    if (!lua_isfunction(L, -1)) {
        printf("WARNING: Callback %s is not a function\\n", func);
        return;
    }
    lua_pop(L, 1);
"""
    
        for call in self.calls:
            print >> swig, LuaCall.luaCallbackRegisterSelector(call, "request") + "\n"
            print >> swig, LuaCall.luaCallbackRegisterSelector(call, "response") + "\n"
        for event in self.events:
            print >> swig, LuaCall.luaCallbackRegisterSelector(event, "event") + "\n"
        
        print >> swig, """
}
%}

%include <typemaps.i>
%include <stdint.i>

%typemap(in, checkfn="lua_isstring") uint8_t[ANY] {
  $1 = (unsigned char*) malloc(lua_strlen(L,$input) / 2);
  convert_string_2_uint8t(lua_tostring(L,$input), $1, lua_strlen(L,$input) / 2);
}

%typemap( freearg ) unsigned char[ANY] {
  free($1);
}

%typemap(in, checkfn="lua_isstring") dsid_t {
  if (lua_strlen(L,$input) != 24) {
    lua_pushstring(L, "String length mismatch for conversion into dsid_t");
    SWIG_fail;
  }
  convert_string_2_uint8t(lua_tostring(L,$input), (uint8_t*) &$1, sizeof(dsid_t));
}


%typemap(out) dsid_t {
    char stringdata[sizeof(dsid_t) * 2 + 1];
    convert_uint8t_2_string($1.id, stringdata, sizeof(dsid_t) * 2);
    lua_pushstring(L, stringdata);
}

%typemap(out) dsid_t* {
    char stringdata[sizeof(dsid_t) * 2 + 1];
    convert_uint8t_2_string((uint8_t*) $1, stringdata, sizeof(dsid_t) * 2);
    lua_pushstring(L, stringdata);
}

%typemap(in,numinputs=0) dsid_t *dsid(dsid_t temp) {
    $1 = &temp;
}

%typemap(in,numinputs=0) (dsid_t *dsid)(uint8_t temp) {
   $1=(dsid_t*)malloc(sizeof(dsid_t));
}

%typemap(argout) dsid_t *dsid {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      char stringdata[sizeof(dsid_t) * 2 + 1];
      convert_uint8t_2_string($1->id, stringdata, sizeof(dsid_t) * 2);
      lua_pushstring(L, stringdata);
   }
   SWIG_arg++;
   free($1);
}

%typemap(in) (dsid_t *device_list, int device_list_size) {
   if (!lua_isnumber(L,$input)) {
       error(L, "Expecting number");
   }
   $2 = lua_tonumber(L, $input);
   if ($2 < 0) {
       error(L, "Positive integer expected");
   }
   $1 = (dsid_t*) malloc(sizeof(dsid_t) * $2);
}

%typemap(argout) (dsid_t *device_list, int device_list_size) {
    char stringdata[sizeof(dsid_t) * 2 + 1];
    int i;
    lua_newtable(L);
    for (i = 0; i < result; i ++) {
        convert_uint8t_2_string( & ((uint8_t *) $1) [i * sizeof(dsid_t)], stringdata, sizeof(dsid_t) * 2);
        lua_pushstring(L, stringdata);
        lua_rawseti(L, -2, i+1); 
    }
    SWIG_arg++;
    free($1);
}

%typemap(in,numinputs=0) bus_state_t *state(bus_state_t temp) {
    $1 = &temp;
}

%typemap(argout) bus_state_t *state {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint8_t * (uint8_t temp) {
   $1 = &temp;
}

%typemap(argout) uint8_t * {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint16_t * (uint16_t temp) {
   $1 = &temp;
}

%typemap(argout) uint16_t * {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint32_t * (uint32_t temp) {
   $1 = &temp;
}

%typemap(argout) uint32_t * {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint8_t *Out(uint8_t temp) {
   $1 = &temp;
}

%typemap(argout) uint8_t *Out {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint16_t *Out(uint16_t temp) {
   $1 = &temp;
}

%typemap(argout) uint16_t *Out {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint32_t *Out(uint32_t temp) {
   $1 = &temp;
}

%typemap(argout) uint32_t *Out {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      lua_pushnumber(L, *$1);
   }
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint8_t *OutName(uint8_t temp) {
$1=(uint8_t*)malloc(20);
}
%typemap(argout) uint8_t *OutName {
   if ($result < 0) {
      lua_pushnil(L);
   } else {
      char stringdata[2 * 20 + 1];
      convert_uint8t_2_string($1, stringdata, 2 * 20);
      lua_pushstring(L, stringdata);
   }
   free($1);
   SWIG_arg++;
}

"""
        if (len(length_enum) > 0):
            s = Template("""%typemap(argout) uint8_t *OutArray${length} {
   if ($$result < 0) {
      lua_pushnil(L);
   } else {
      char stringdata[2 * ${length} + 1];
      convert_uint8t_2_string($$1, stringdata, 2 * ${length});
      lua_pushstring(L, stringdata);
   }
   free($$1);
   SWIG_arg++;
}

%typemap(in,numinputs=0) uint8_t *OutArray${length}(uint8_t temp) {
    $$1=(uint8_t*)malloc(${length});
}
""")
            print >> swig, "".join(map(lambda x: s.substitute(length = str(x)), length_enum.itervalues()))

        for call in self.calls:
            print >> swig, PyCall.call_signature_swig(call) + ";"
        
        print >> swig, """void DsmApiRegisterLuaCallback(DsmApiHandle_t handle, container_type_t MessageType, int MessageId, int ModifierId, const char *func, lua_State* L);
void DsmApiRegisterLuaBusStateCallback(DsmApiHandle_t handle, const char *func, lua_State* L);
void DsmApiRegisterLuaBusChangeCallback(DsmApiHandle_t handle, const char *func, lua_State* L);

%include "dsm-api-const.h"
%include "dsm-api.h"
"""
        swig.close()

