# This file is part of the digitalSTROM dsm-api library.
#
# Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland
#
# digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# digitalSTROM dsm-api library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.

from SwigGenerator import *

class PyParameter(SwigParameter):
    @classmethod
    def getDoc(cls, parameter):
        return parameter.name + " -- " + parameter.description

    @classmethod
    def getPyTypeString(cls, parameter):
        if parameter.type_size == 8:
            if (parameter.type_count == 1):
                return "B"
            else:
                return "[" + "B" * parameter.type_count + "]"
        elif parameter.type_size == 16:
            return "H"
        elif parameter.type_size == 32:
            return "I"
        
    @classmethod
    def getPyCall(cls, parameter):
        if (parameter.type_count == 1):
            return parameter.name
        else:
            if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
                return ", ".join(map(lambda x: "DSID.id[" + str(x) + "]", range(parameter.type_count)))
            else:
                return ", ".join(map(lambda x: parameter.name + "[" + str(x) + "]", range(parameter.type_count)))

class PyCall(SwigCall):
    @classmethod
    def py_callback_implementation(cls, call, type):
        s = Template("""void ${name}_${type}_py_callback(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId${parameter_decl})
{
    PyObject *arglist;
    PyObject *result;
    PyObject* python_callback = (PyObject*)arg;
    
    arglist = Py_BuildValue("(B${type_string})", error_code, sourceId.id[0], sourceId.id[1], sourceId.id[2], sourceId.id[3], 
                                                sourceId.id[4], sourceId.id[5], sourceId.id[6], sourceId.id[7], 
                                                sourceId.id[8], sourceId.id[9], sourceId.id[10], sourceId.id[11], 
                                                destinationId.id[0], destinationId.id[1], destinationId.id[2], destinationId.id[3], 
                                                destinationId.id[4], destinationId.id[5], destinationId.id[6], destinationId.id[7], 
                                                destinationId.id[8], destinationId.id[9], destinationId.id[10], destinationId.id[11]
                                                ${param_call});
    
    SWIG_PYTHON_THREAD_BEGIN_BLOCK;
    
    result = PyEval_CallObject(python_callback, arglist);
    
    Py_XDECREF(arglist);
    Py_XDECREF(result);
    
    SWIG_PYTHON_THREAD_END_BLOCK;
}
""")
        if (type == "request"):
            parameters = call.request_params
        else:
            parameters = call.response_params
        
        parameter_decl = ""
        param_call = ""
        type_string = ("[" + "B" * 12 + "]") * 2
        if parameters:
            parameter_decl = ", " + ", ".join(map(lambda x: PyParameter.signature(x), parameters))
            param_call = ", " + ", ".join(map(lambda x: PyParameter.getPyCall(x), parameters))
            type_string += "".join(map(lambda x: PyParameter.getPyTypeString(x), parameters))
            
        return s.substitute(name = call.name,
                            type = type,
                            parameter_decl = parameter_decl,
                            type_string = type_string,
                            param_call = param_call)
    
    @classmethod
    def pyCallbackRegisterSelector(cls, call, call_type):
        s = Template("""    if (MessageType == ${type_const} && MessageId == ${message_id} && ModifierId == ${modifier_id})
    {
        DsmApiRegisterCallback(handle, MessageType, MessageId, ModifierId, &${name}_${type}_py_callback, pyfunc);
        Py_INCREF(pyfunc);
        return;
    }""")
        
        if call_type == "response":
            type_const = "DS485_CONTAINER_RESPONSE"
        elif call_type == "request":
            type_const = "DS485_CONTAINER_REQUEST"
        else:
            type_const = "DS485_CONTAINER_EVENT"
            
        return s.substitute(message_id = str(call.message_id),
                            modifier_id = str(call.modifier_id),
                            name = call.name,
                            type = call_type,
                            type_const = type_const)

    @classmethod
    def py_docstring(cls, call):
        s = Template("""%feature("autodoc", "${name}(${inParameters}) -> (${outParameters})") ${name};
%feature("docstring") "$description

This call has a timeout of $timeout seconds.

Arguments:
$request

Return values:
$response" """)
        return s.substitute(name = call.name,
                            inParameters = ", ".join(["handle", "dSMdSID"] + map(lambda x: x.name, call.request_params)),
                            outParameters = ", ".join(["error"] + map(lambda x: x.name, call.response_params)),
                            description = call.description,
                            timeout = call.timeout,
                            request = "\n".join(["handle -- Handle to the library instance", "dSMdSID -- Destination of the call"] + map(lambda x: PyParameter.getDoc(x), call.request_params)),
                            response = "\n".join(["error -- Errorcode: 0: success (return parameters are valid); <0: Library error; >0: dSM error code"] + map(lambda x: PyParameter.getDoc(x), call.response_params)))


class PyGenerator:
    def __init__(self, license, calls, events):
        self.calls = calls
        self.events = events
        self.license = license
        
    def swig(self, swig_decl_file, length_enum):
        swig = open(swig_decl_file, 'w')
        print >> swig, self.license
        print >> swig, """%module dsmapi2
%{
#include "dsm-api-const.h"
#include <digitalSTROM/ds.h>
static int convert_uint8_t(PyObject *input, uint8_t *ptr, int size) {
  int i;
  if (!PySequence_Check(input)) {
      PyErr_SetString(PyExc_TypeError,"Expecting a sequence");
      return 0;
  }
  if (PyObject_Length(input) != size) {
      PyErr_SetString(PyExc_ValueError,"Sequence size mismatch");
      return 0;
  }
  for (i =0; i < size; i++) {
      PyObject *o = PySequence_GetItem(input,i);
      if (!PyInt_Check(o)) {
         Py_XDECREF(o);
         PyErr_SetString(PyExc_ValueError,"Expecting a sequence of integers");
         return 0;
      }
      ptr[i] = PyInt_AsLong(o);
      Py_DECREF(o);
  }
  return 1;
}
static int convert_string(PyObject *input, uint8_t *ptr, int size) {
  int i;
  if (!PySequence_Check(input)) {
      PyErr_SetString(PyExc_TypeError,"Expecting a sequence");
      return 0;
  }
  if (PyObject_Length(input) > size) {
      PyErr_SetString(PyExc_ValueError,"Sequence size mismatch");
      return 0;
  }
  if (!PyString_Check(input))
  {
       PyErr_SetString(PyExc_ValueError,"Expecting a string");
     return 0;
  }
  memcpy(ptr, PyString_AsString(input), PyString_Size(input));
  memset(ptr + PyString_Size(input), 0, 20-PyString_Size(input));
  return 1;
}

static void convert_uint8_array_to_pylist(PyObject **result, uint8_t *data, uint8_t length)
{
    int i;
    PyObject *o, *o2, *o3;
    o = PyList_New(length);
    for (i = 0; i < length; ++i) {
        PyObject *item = PyInt_FromLong(data[i]);
        PyList_SetItem(o, i, item);
    }
    if ((!*result) || (*result == Py_None)) {
        *result = o;
    } else {
        if (!PyTuple_Check(*result)) {
            PyObject *o2 = *result;
            *result = PyTuple_New(1);
            PyTuple_SetItem(*result, 0, o2);
        }
        o3 = PyTuple_New(1);
        PyTuple_SetItem(o3, 0, o);
        o2 = *result;
        *result = PySequence_Concat(o2, o3);
        Py_DECREF(o2);
        Py_DECREF(o3);
    }
}

static void convert_uint8_to_pylong(PyObject **result, uint8_t *data)
{
    PyObject *o, *o2, *o3;
    o = PyInt_FromLong(*data);
    if ((!*result) || (*result == Py_None)) {
        *result = o;
    } else {
        if (!PyTuple_Check(*result)) {
            PyObject *o2 = *result;
            *result = PyTuple_New(1);
            PyTuple_SetItem(*result, 0, o2);
        }
        o3 = PyTuple_New(1);
        PyTuple_SetItem(o3, 0, o);
        o2 = *result;
        *result = PySequence_Concat(o2, o3);
        Py_DECREF(o2);
        Py_DECREF(o3);
    }
}

static void convert_uint16_to_pylong(PyObject **result, uint16_t *data)
{
    PyObject *o, *o2, *o3;
    o = PyInt_FromLong(*data);
    if ((!*result) || (*result == Py_None)) {
        *result = o;
    } else {
        if (!PyTuple_Check(*result)) {
            PyObject *o2 = *result;
            *result = PyTuple_New(1);
            PyTuple_SetItem(*result, 0, o2);
        }
        o3 = PyTuple_New(1);
        PyTuple_SetItem(o3, 0, o);
        o2 = *result;
        *result = PySequence_Concat(o2, o3);
        Py_DECREF(o2);
        Py_DECREF(o3);
    }
}

static void convert_uint32_to_pylong(PyObject **result, uint32_t *data)
{
    PyObject *o, *o2, *o3;
    o = PyLong_FromUnsignedLong(*data);
    if ((!*result) || (*result == Py_None)) {
        *result = o;
    } else {
        if (!PyTuple_Check(*result)) {
            PyObject *o2 = *result;
            *result = PyTuple_New(1);
            PyTuple_SetItem(*result, 0, o2);
        }
        o3 = PyTuple_New(1);
        PyTuple_SetItem(o3, 0, o);
        o2 = *result;
        *result = PySequence_Concat(o2, o3);
        Py_DECREF(o2);
        Py_DECREF(o3);
    }
}

static void convert_uint8_attay_to_pystring(PyObject **result, uint8_t *data)
{
    PyObject *o, *o2, *o3;
    o = PyString_FromString(data);
    if ((!*result) || (*result == Py_None)) {
        *result = o;
    } else {
        if (!PyTuple_Check(*result)) {
            PyObject *o2 = *result;
            *result = PyTuple_New(1);
            PyTuple_SetItem(*result, 0, o2);
        }
        o3 = PyTuple_New(1);
        PyTuple_SetItem(o3, 0, o);
        o2 = *result;
        *result = PySequence_Concat(o2, o3);
        Py_DECREF(o2);
        Py_DECREF(o3);
    }
}

/* Includes the header in the wrapper code */
#include "dsm-api.h"
"""
        for call in self.calls:
            print >> swig, PyCall.py_callback_implementation(call, "request")
            print >> swig, PyCall.py_callback_implementation(call, "response")
        for event in self.events:
            print >> swig, PyCall.py_callback_implementation(event, "event")
            
        print >> swig, """
static void BusChange_callback(void *arg, dsid_t *id, int flag)
{
    PyObject *arglist;
    PyObject *result;
    PyObject* python_callback = (PyObject*)arg;
    
    arglist = Py_BuildValue("([BBBBBBBBBBBB]B)", id->id[0], id->id[1], id->id[2], id->id[3], 
                                                id->id[4], id->id[5], id->id[6], id->id[7], 
                                                id->id[8], id->id[9], id->id[10], id->id[11],
                                                flag);
    
    SWIG_PYTHON_THREAD_BEGIN_BLOCK;
    
    result = PyEval_CallObject(python_callback, arglist);
    
    Py_XDECREF(arglist);
    Py_XDECREF(result);
    
    SWIG_PYTHON_THREAD_END_BLOCK;
}

void DsmApiRegisterPyBusChangeCallback(DsmApiHandle_t handle, PyObject *pyfunc)
{
    if (!PyCallable_Check(pyfunc))
    {
        DsmApiRegisterBusChangeCallback(handle, 0, 0);
        return;
    }
    
    DsmApiRegisterBusChangeCallback(handle, &BusChange_callback, pyfunc);
}

static void BusState_callback(void *arg, bus_state_t state)
{
    PyObject *arglist;
    PyObject *result;
    PyObject* python_callback = (PyObject*)arg;
    
    arglist = Py_BuildValue("(B)", state);
    
    SWIG_PYTHON_THREAD_BEGIN_BLOCK;
    
    result = PyEval_CallObject(python_callback, arglist);
    
    Py_XDECREF(arglist);
    Py_XDECREF(result);
    
    SWIG_PYTHON_THREAD_END_BLOCK;
}

void DsmApiRegisterPyBusStateCallback(DsmApiHandle_t handle, PyObject *pyfunc)
{
    if (!PyCallable_Check(pyfunc))
    {
        DsmApiRegisterBusStateCallback(handle, 0, 0);
        return;
    }
    
    DsmApiRegisterBusStateCallback(handle, &BusState_callback, pyfunc);
}
void DsmApiRegisterPyCallback(DsmApiHandle_t handle, container_type_t MessageType, int MessageId, int ModifierId, PyObject *pyfunc)
{
    if (!PyCallable_Check(pyfunc))
    {
        DsmApiRegisterCallback(handle, MessageType, MessageId, ModifierId, 0, 0);
        return;
    }
"""
    
        for call in self.calls:
            print >> swig, PyCall.pyCallbackRegisterSelector(call, "request") + "\n"
            print >> swig, PyCall.pyCallbackRegisterSelector(call, "response") + "\n"
        for event in self.events:
            print >> swig, PyCall.pyCallbackRegisterSelector(event, "event") + "\n"
        
        print >> swig, """
}
%}

%typemap(in,numinputs=0) uint8_t *Out(uint8_t temp) {
    $1 = &temp;
}

%typemap(in) uint8_t {
    $1 = PyInt_AsLong($input);
}

%typemap(argout) uint8_t *Out {
    convert_uint8_to_pylong(&$result, $1);
}

%typemap(in) uint16_t {
    $1 = PyInt_AsLong($input);
}

%typemap(in,numinputs=0) uint16_t *Out(uint16_t temp) {
    $1 = &temp;
}

%typemap(argout) uint16_t *Out {
    convert_uint16_to_pylong(&$result, $1);
}

%typemap(in) uint32_t {
    $1 = PyLong_AsUnsignedLong($input);
}

%typemap(in,numinputs=0) uint32_t *Out(uint32_t temp) {
    $1 = &temp;
}

%typemap(argout) uint32_t *Out {
    convert_uint32_to_pylong(&$result, $1);
}

%typemap(in,numinputs=0) (dsid_t *dsid)(uint8_t temp) {
    $1=(dsid_t*)malloc(sizeof(dsid_t));
}

%typemap(in) dsid_t {
    if (!convert_uint8_t($input, (uint8_t*)&$1, 12)) {
        return NULL;
    }
}

%typemap(argout) (dsid_t *dsid) {
    convert_uint8_array_to_pylist(&$result, (uint8_t*)$1, 12);
    free($1);
}

%typemap(in) (dsid_t *device_list, int device_list_size) {
    if (!PyInt_Check($input)) {
       PyErr_SetString(PyExc_ValueError, "Expecting an integer");
       return NULL;
   }
   $2 = PyInt_AsLong($input);
   if ($2 < 0) {
       PyErr_SetString(PyExc_ValueError, "Positive integer expected");
       return NULL;
   }
   $1 = (dsid_t*)malloc(sizeof(dsid_t) * $2);
}

%typemap(argout) (dsid_t *device_list, int device_list_size) {
    PyObject *o, *o2, *o3;
    int i = 0;
    int num_results = PyInt_AsLong($result);
    for (i = 0; i < num_results; i++) {
        convert_uint8_array_to_pylist(&$result, (uint8_t*)((uint8_t*)$1 + (i * sizeof(dsid_t))), sizeof(dsid_t));
    }
    free($1);
}

%typemap(in,numinputs=0) bus_state_t *state(bus_state_t temp) {
    $1 = &temp;
}

%typemap(argout) bus_state_t *state {
    convert_uint8_to_pylong(&$result, (uint8_t*)$1);
}

%typemap(in) uint8_t *InName {
    $1 = (uint8_t*)malloc(20);
   if (!convert_string($input, $1, 20)) {
      return NULL;
   }
}

%typemap(freearg) uint8_t *InName {
   free($1);
}

%typemap(in,numinputs=0) uint8_t *OutName(uint8_t temp) {
   $1=(uint8_t*)malloc(20);
}

%typemap(argout) uint8_t *OutName {
    convert_uint8_attay_to_pystring(&$result, $1);
    free($1);
}
"""
        if (len(length_enum) > 0):
            s = Template("""%typemap(in,numinputs=0) uint8_t *OutArray${length}(uint8_t temp) {
    $$1=(uint8_t*)malloc(${length});
}

%typemap(argout) uint8_t *OutArray${length} {
    convert_uint8_array_to_pylist(&$$result, $$1, ${length});
    free($$1);
}

%typemap(in) uint8_t *InArray${length} {
    $$1 = (uint8_t*)malloc(${length});
   if (!convert_uint8_t($$input, $$1, ${length})) {
      return NULL;
   }
}

%typemap(freearg) uint8_t *InArray${length} {
   free($$1);
}
""")
            print >> swig, "".join(map(lambda x: s.substitute(length = str(x)), length_enum.itervalues()))

        for call in self.calls:
            print >> swig, PyCall.py_docstring(call)
            print >> swig, PyCall.call_signature_swig(call) + ";"
        
        print >> swig, """%feature("autodoc", "DsmApiInitialize() -> (handle)") DsmApiInitialize;
%feature("docstring") "
Initialize the internal memory structures of the library.

Return values: 
handle -- a handle to the library instance"
DsmApiHandle_t DsmApiInitialize(void);

%feature("autodoc", "DsmApiOpen(handle, connection, flags) -> (error)") DsmApiOpen;
%feature("docstring") "Open a new connection to the dS485 bus.

Arguments:
handle -- Handle to the library instance
connection  -- string containing the connection method: 
                      - serial:///dev/ttyUSB0 (only when linked to the GPL version of the ds485-client library)
                      - tcp://localhost:8442
flags -- pass 0

Return values:
error -- zero on success, negative on failure"
int DsmApiOpen(DsmApiHandle_t handle, const char *connection, int flags);

%feature("autodoc", "DsmApiGetOwnDSID(handle) -> (error, dsid)") DsmApiGetOwnDSID;
%feature("docstring") "Get the dSID of the daeomon/socket interface.

Arguments:
handle -- Handle to the library instance

Return values:
error -- 0 on success; negative on failure
dsid -- own dSID"
int DsmApiGetOwnDSID(DsmApiHandle_t handle, dsid_t *dsid);

%feature("autodoc", "DsmApiGetBusMembers(handle, device_list_size) -> (error, device_list)") DsmApiGetBusMembers;
%feature("docstring") "Request a list of the current bus members.

Arguments:
handle -- Handle to the library instance
device_list_size -- device_list has space for this many devices.

Return values:
error -- the number of devices in device_list; negative if an error occured
device_list -- List of devices. Preallocate a buffer of size device_list_size * sizeof(dsid_t)"
int DsmApiGetBusMembers(DsmApiHandle_t handle, dsid_t *device_list, int device_list_size);

%feature("autodoc", "DsmApiGetBusState(handle) -> (error, state)") DsmApiGetBusState;
%feature("docstring") "Get the current state of the dS485 bus.

Arguments:
handle -- Handle to the library instance

Return values:
error -- 0 on success; negative on failure
state -- the bus state"
int DsmApiGetBusState(DsmApiHandle_t handle, bus_state_t *state);

%feature("autodoc", "DsmApiClose(handle) -> (error)") DsmApiClose;
%feature("docstring") "Close the connection to the dS485 bus.

Arguments:
handle -- Handle to the library instance

Return values:
error -- 0 on success; negative on failure"
int DsmApiClose(DsmApiHandle_t handle);

%feature("autodoc", "DsmApiCleanup(handle) -> (error)") DsmApiCleanup;
%feature("docstring") "Free instance memory.

Arguments:
handle -- Handle to the library instance

Return values:
error -- 0 on success; negative on failure"
int DsmApiCleanup(DsmApiHandle_t handle);

%feature("autodoc", "DsmApiRegisterPyCallback(handle, MessageType, MessageId, ModifierId, pyfunc) -> (error)") DsmApiRegisterPyCallback;
%feature("docstring") "Register callback functions for dsm-api asynchronous responses, requests and events.

Register NULL to delete a callback.

Arguments:
handle -- Handle to the library instance
MessageType -- Register callback function for this MessageType
MessageId -- Register callback function for this MessageId
ModifierId -- Register the callback function for thei ModifierId
pyfunc -- the callback function

Return values:
callback -- the previously registered callback function for this MessageId/ModifierId combination"
void DsmApiRegisterPyCallback(DsmApiHandle_t handle, container_type_t MessageType, int MessageId, int ModifierId, PyObject *pyfunc);

%feature("autodoc", "DsmApiRegisterPyBusChangeCallback(handle, pyfunc) -> (error)") DsmApiRegisterPyBusChangeCallback;
%feature("docstring") "Register a callback function for BusChange events.

Register NULL to delete the callback.

Arguments:
handle -- Handle to the library instance
pyfunc -- the callback function

Return values:
callback -- the previously registered callback function"
void DsmApiRegisterPyBusChangeCallback(DsmApiHandle_t handle, PyObject *pyfunc);

%feature("autodoc", "DsmApiRegisterPyBusStateCallback(handle, pyfunc) -> (error)") DsmApiRegisterPyBusStateCallback;
%feature("docstring") "Register a callback function for BusState events.

Register NULL to delete the callback.

Arguments:
handle -- Handle to the library instance
pyfunc -- the callback function

Return values:
callback -- the previously registered callback function"
void DsmApiRegisterPyBusStateCallback(DsmApiHandle_t handle, PyObject *pyfunc);

%feature("autodoc", "DeviceStatisticsRequest_get_sync(handle, dSMdSID, DeviceId, ParamOffset, timeout) -> (error, Value)") DeviceStatisticsRequest_get_sync;
%feature("docstring") "Requests statistics data of a device. Synchronous implementation of #DeviceStatisticsRequest_get.

ParamOffset 0xFE (Reset) and 0xFF (All values) are not allowed, as they would result in no or multiple response events.

Arguments:
handle -- Handle to the library instance
dSMdSID -- Destination of the call
DeviceId --
ParamOffset -- 
timeout -- time-out of the synchronous call (in Seconds)

Return values:
error -- Errorcode: 0: success (return parameters are valid); <0: Library error; >0: dSM error code
Value -- Statistics value (only valid if #error == 0)"
int DeviceStatisticsRequest_get_sync(DsmApiHandle_t handle, dsid_t dsid, uint16_t DeviceId, uint8_t ParamOffset, int timeout, uint16_t *Out);

%feature("autodoc", "DeviceConfig_get_sync_8(handle, dSMdSID, DeviceId, ConfigClass, ConfigIndex, timeout) -> (error, Value)") DeviceConfig_get_sync_8;
%feature("docstring") "Read a configuration parameter of a specific digitalSTROM device. Synchronous implementation of #DeviceConfigRequest_get.

Arguments:
handle -- Handle to the library instance
dSMdSID -- Destination of the call
DeviceId --
ConfigClass --
ConfigIndex --
timeout -- time-out of the synchronous call (in Seconds)

Return values:
error -- Errorcode: 0: success (return parameters are valid); <0: Library error; >0: dSM error code
Value -- Parameter value (only valid if return == 0)"
int DeviceConfig_get_sync_8(DsmApiHandle_t handle, dsid_t dsid, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint8_t *Out);

%feature("autodoc", "DeviceConfig_get_sync_16(handle, dSMdSID, DeviceId, ConfigClass, ConfigIndex, timeout) -> (error, Value)") DeviceConfig_get_sync_16;
%feature("docstring") "Read two configuration parameters of a specific digitalSTROM device simultaneously. Synchronous implementation of #DeviceConfigRequest_get.

Arguments:
handle -- Handle to the library instance
dSMdSID -- Destination of the call
DeviceId --
ConfigClass --
ConfigIndex --
timeout -- time-out of the synchronous call (in Seconds)

Return values:
error -- Errorcode: 0: success (return parameters are valid); <0: Library error; >0: dSM error code
Value -- Parameter values low byte: ConfigIndex & 0xFE; high byte: ConfigIndex & 0xFE + 1 (only valid if return == 0)"
int DeviceConfig_get_sync_16(DsmApiHandle_t handle, dsid_t dsid, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint16_t *Out);

%feature("autodoc", "TestTransmissionQuality_get_sync(handle, dSMdSID, DeviceId, ConfigClass, ConfigIndex, timeout) -> (error, DownstreamQuality, UpstreamQuality)") TestTransmissionQuality_get_sync;
%feature("docstring") "Query transmission quality. Synchronous implementation of #TestTransmissionQuality_get.

Arguments:
handle -- Handle to the library instance
dSMdSID -- Destination of the call
DeviceId --
timeout -- time-out of the synchronous call (in Seconds)

Return values:
error -- Errorcode: 0: success (return parameters are valid); <0: Library error; >0: dSM error code
DownstreamQuality -- Downstream quality (only valid if return == 0)
UpstreamQuality -- Upstream quality (only valid if return == 0)"
int TestTransmissionQuality_get_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, int timeout, uint8_t *Out, uint16_t *Out);

typedef enum {
    DS485_CONTAINER_REQUEST = 1,
    DS485_CONTAINER_RESPONSE,
    DS485_CONTAINER_EVENT
} container_type_t;

typedef void *DsmApiHandle_t;
%include "dsm-api-const.h"
"""
        swig.close()
        