# This file is part of the digitalSTROM dsm-api library.
#
# Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland
#
# digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# digitalSTROM dsm-api library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.

from CGenerator import *

class SwigParameter(Parameter):
    @classmethod
    def signature_swig(cls, parameter):
        if parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "dsid_t InDSID"
        
        out = "uint" + str(parameter.type_size) + "_t "
        if parameter.name == "Name":
            out += "*InName"
        elif (parameter.type_count > 1):
            out += "*InArray" + str(parameter.type_count)
        else:
            out += parameter.name
        return out
        
    @classmethod
    def declare_by_reference_swig(cls, parameter):
        if (parameter.type_count == 1):
            return "uint" + str(parameter.type_size) + "_t *Out"
        elif parameter.name == "Name":
            return "uint" + str(parameter.type_size) + "_t *OutName"
        elif parameter.name == "DSID" and parameter.type_size == 8 and parameter.type_count == 12:
            return "dsid_t *dsid"
        else:
            return "uint" + str(parameter.type_size) + "_t *OutArray" + str(parameter.type_count)
        
class SwigCall(Call):
    @classmethod
    def call_signature_swig(cls, call):
        out = "int " + call.name + "(DsmApiHandle_t handle, dsid_t dSMdSID"
        if call.request_params:
            out += ", " + ", ".join(map(lambda x: SwigParameter.signature_swig(x), call.request_params))
        if call.response_params:
            out += ", " + ", ".join(map(lambda x: SwigParameter.declare_by_reference_swig(x), call.response_params))
        out += ")"
        return out